# UpdateManager.py
# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-
import os
import apt_pkg
import sys
import time
import shutil
import dbus
import logging
import dbus.service
import subprocess
import traceback
from apt import Cache
from gettext import gettext as _
from apt.debfile import DebPackage
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)

from .Core.errors import *
from .Core.enums import *
from .Core.MyCache import MyCache
from .UpdateManagerDbus import UpdateManagerDbusController,UpdateManagerDbusControllerUtils,UPDATER_DBUS_INTERFACE,UPDATER_DBUS_PATH,UPDATER_DBUS_SERVICE
from .Core.UpdateList import UpdateList
from .Core.PluginManager import *
from .backend import InstallBackend,get_backend
from .backend import InhibitShutdownLock
from .Core.Database import Sqlite3Server
from .Core.loop import mainloop
from .Core.DataAcquisition import UpdateMsgCollector

from SystemUpdater.Core.LocalUpgradeData import LocalUpgradeDataList
from SystemUpdater.Core.UpdaterConfigParser import UpgradeConfig
from SystemUpdater.Core.JsonConfigParser import JsonConfig as json_config
from SystemUpdater.Core.DpkgInstallProgress import LogInstallProgress
from SystemUpdater.Core.utils import deb_verify,PolicyKit_Authority,get_proc_from_dbus_name,check_source,get_dist,ApplicationSecInstall

class UpdateManager():
    OPENKYLIN_DISTTRIBUTOR = "Openkylin"
    BACKEND_PKG_NAME = 'kylin-system-updater'
    FRONTEND_PKG_NAME = "kylin-update-frontend"
    SYSTEM_UPDATE_GROUPS = "kylin-update-desktop-system"
    GROUPS_PKG_NAME = 'kylin-update-desktop-config'
    SOURCES_UPDATE_NAME = "kylin-software-properties"
    APTD_PKG_NAME = "aptdaemon"
    RUN_UNATTENDED_UPGRADE = '/var/run/unattended-upgrades.pid'
    IMPORTANT_LIST_PATH = '/var/lib/kylin-software-properties/template/important.list'
    RETRY_LIMIT_NUM = 2
    
    def __init__(self,options):
        try:
            self.options = options
            self.cache = None
            self.update_list = None
            self.init_config_aptdeamon = False
            self.aptd_lang_switch = False
            self.retry_limit = self.RETRY_LIMIT_NUM
            self.now_working = InstallBackend.ACTION_DEFUALT_STATUS
            self.lask_working = InstallBackend.ACTION_DEFUALT_STATUS
            
            self._check_exception_status()
            self.dist = get_dist()
            self.bus = dbus.SystemBus()
            #dbus
            self.dbusController = self._setup_dbus(self.bus)
            self.dbusControllerUtils = self._setup_dbus_utils(self.bus)
            #config
            self.configs_uncover = UpgradeConfig(defaults_dir="system-updater-defaults.conf")
            self.configs_cover = UpgradeConfig(name = "system-updater-coverable.conf")
            self.uuconfigs = UpgradeConfig(datadir = "/var/lib/unattended-upgrades/", name = "unattended-upgrades-policy.conf")
            #数据采集器
            self.collector = UpdateMsgCollector(self)
            #连接数据库
            self.sqlite3_server = Sqlite3Server(self)
            self.apt_p2p_config = AptP2pConfigManager()
            self.local_meta = LocalUpgradeDataList()
            json_config()
            self.dbusController.make_aptdeamon_restart()
            self.delay_shutdown = DelayShutdownLock(self)
            self.refresh_cache()
        except Exception as e:
            logging.error(e)
            traceback.print_exc()

    def run(self):
        """Start the daemon and listen for calls."""
        logging.info("Waiting for calls...")
        try:
            mainloop.run()
        except KeyboardInterrupt:
            self.dbusController.Quit(None)

    def _check_exception_status(self):
        logging.info("Check: Abonormal Device Startup conditions...")
        InhibitShutdownLock._check_unlock_sleep()

    @property
    def is_openkylin(self):
        return self.dist == self.OPENKYLIN_DISTTRIBUTOR

    #进行清空所有下载的文件
    def start_clean(self):
        clean_backend = get_backend(self, InstallBackend.ACTION_CLEAN)
        clean_backend.start()

    #进行修复破损的包的操作 apt install -f 
    def start_fix_broken(self):
        fix_backend = get_backend(self, InstallBackend.ACTION_FIX_BROKEN)
        fix_backend.start()
    
    #进行 dpkg --configure
    def start_fix_incomplete(self):
        fix_backend = get_backend(self, InstallBackend.ACTION_FIX_INCOMPLETE)
        fix_backend.start()
    
    #计算依赖关系模块
    def start_resolver_depends(self,upgrade_mode = InstallBackend.MODE_DEFAULT_STATUS,push_content = []):
        try:
            if upgrade_mode == InstallBackend.MODE_INSTALL_PARTIAL:
                push_content += self.local_meta.will_upgrade_content

            self.check_config_aptdeamon()
            resolver_backend = get_backend(self, InstallBackend.ACTION_CHECK_RESOLVER,upgrade_mode)
            resolver_backend.start_resolver(push_content)
        except Exception as e:
            logging.error(e)

    #进行升级的操作
    def start_install(self,upgrade_mode = InstallBackend.MODE_DEFAULT_STATUS,push_content = []):
        try:
            action = InstallBackend.ACTION_DEFUALT_STATUS
            if self.local_meta.need_poweroff_install == True:
                action = InstallBackend.ACTION_DOWNLOADONLY
            else:
                action = InstallBackend.ACTION_INSTALL

            install_backend = get_backend(self,action,upgrade_mode)
            install_backend.start(push_content)
        except Exception as e:
            logging.error(e)

    #自我升级
    def start_selfupgrade(self,push_content):
        install_backend = get_backend(self,InstallBackend.ACTION_SLEF_UPGRADE,InstallBackend.MODE_DEFAULT_STATUS)
        install_backend.start(push_content)

    #关机安装
    def start_poweroff(self,by_trigger="dbus"):
        upgrade_mode = InstallBackend.MODE_SHUTDOWN_BY_DBUS
        if bool(self.local_meta.check_install_required()) == True:
            if self.local_meta.trigger_by_reboot == True and by_trigger != 'dbus':
                upgrade_mode = InstallBackend.MODE_SHUTDOWN_BY_USER
                self.local_meta.finished_status = self._check_finished_status()

            install_backend = get_backend(self, InstallBackend.ACTION_INSTALL_SHUTDOWN,upgrade_mode)
            install_backend.start(self.local_meta.will_upgrade_content)

    #关机安装
    def start_task_for_auto(self,action):
        task_backend = get_backend(self, action)
        task_backend.start()

    #进行更新的操作
    def start_update(self,update_mode = InstallBackend.MODE_UPDATE_ALL):
        try:
            self._check_network()

            self._update_important()

            self.start_update_backend(update_mode = update_mode)
        except UpdateBaseError as excep:
            self.dbusController.UpdateDetectFinishedNext(False,[''],excep.code_strnum,excep.header,excep.desc)
        except UpdateProgressExit:
            pass
        except Exception as e:
            logging.error(e)
            traceback.print_exc()
            self.dbusController.UpdateDetectFinishedNext(False,[''],get_error_code_from_enum(ERROR_PROGRAM_EXCEPTION),\
                                                                get_error_string_from_enum(ERROR_PROGRAM_EXCEPTION),\
                                                                get_error_description_from_enum(ERROR_PROGRAM_EXCEPTION))

    def start_update_backend(self,update_mode = InstallBackend.MODE_UPDATE_ALL):
        self.check_config_aptdeamon()
        update_backend = get_backend(self, InstallBackend.ACTION_UPDATE,update_mode)
        update_backend.start()

    def start_available(self):
        try:
            self.configs_cover.reRead()

            self.retry_limit = self.RETRY_LIMIT_NUM

            json_config().read()

            self.refresh_cache()

            self.local_meta.refresh()

            self.update_list = UpdateList(self)

            # self._check_failed_packages(self.cache)

            self._check_system_incomplete(self.cache)
            
            self._check_system_broken(self.cache)
            
            self._check_config_upgrade(self.cache)

            self._check_self_upgrade(self.cache)

            self.update_list.update_kylin(self.cache,self._get_important_data(),self.is_openkylin)

            self._update_last_install()

            self.dbusController.UpdateDetectFinishedNext(True,self.local_meta.get_push(),'','','')

            self.collector.update_process('finish-updatedetect',True,self.local_meta.get_push(),'','','')

            if not self.local_meta.get_push():
                logging.info("The software on this computer is up to date, try to refresh system version ...")
                self.sqlite3_server._repair_system_version()

        except UpdateBaseError as excep:
            self.dbusController.UpdateDetectFinishedNext(False,[''],excep.code_strnum,excep.header,excep.desc)

            self.collector.update_process('finish-updatedetect',False,self.local_meta.get_push(),excep.code_strnum,excep.header,excep.desc)
        except UpdateProgressExit:
            pass
        except Exception as e:
            logging.error(e)
            traceback.print_exc()
            self.dbusController.UpdateDetectFinishedNext(False,[''],get_error_code_from_enum(ERROR_PROGRAM_EXCEPTION),\
                                                                get_error_string_from_enum(ERROR_PROGRAM_EXCEPTION),\
                                                                get_error_description_from_enum(ERROR_PROGRAM_EXCEPTION))

    def refresh_cache(self):
        try:
            self._reload_options_config()
            if self.cache is None:
                self.cache = MyCache(None)
            else:
                self.cache.open(None)
                self.cache._initDepCache()
            logging.info("Refresh: Cache for System Update Successfully...")
        except AssertionError:
            raise UpdateBaseError(ERROR_SOFTWARE_INDEX_RROKEN)
        except SystemError as e:
            logging.error(str(e))
            raise UpdateBaseError(ERROR_NOT_INIT_PACKAGESINFIO)

    def _update_last_install(self):
        last_status = self.configs_uncover.getWithDefault("LastUpdateInfo",\
                            "finished_status",self.local_meta.SHUTDOWN_INSTALL_SUCCESS)

        if last_status == self.local_meta.SHUTDOWN_INSTALL_SUCCESS or last_status == '':
            pass
        else:
            last_version = self.configs_uncover.getWithDefault("LastUpdateInfo","update_version",'')
            update_content = self.configs_uncover.getlist("LastUpdateInfo","update_content")
            now_version = json_config().getWithDefault("version",default = '')
            now_push = self.local_meta.get_push()
            # 比对版本号不同或者更新内容不相同 则认为认为存在新的推送
            if now_version != last_version or update_content != now_push:
                self.configs_uncover.setValue("LastUpdateInfo","finished_status",\
                                                self.local_meta.SHUTDOWN_INSTALL_SUCCESS)
                self.configs_uncover.setValue("LastUpdateInfo","update_version",'')
                self.configs_uncover.setListValue("LastUpdateInfo","update_content",[])

    def _check_network(self):
        if self.options.no_check_network is False and check_source() == False:
            self.dbusController.UpdateDetectStatusChanged(5,_("Checking network connection"))
            try:
                obj = self.bus.get_object("org.freedesktop.NetworkManager","/org/freedesktop/NetworkManager")
                interface = dbus.Interface(obj, "org.freedesktop.DBus.Properties")
                retval = interface.Get("org.freedesktop.NetworkManager","Connectivity",timeout=0.5)
            except Exception:
                retval = 4
            #1 表示没有网卡可以使用
            if retval == 1:
                raise UpdateBaseError(ERROR_NETWORK_FAILED)

    #更新important.list的本次升级的列表
    def _update_important(self):
        def _update_important_reply(retval):
            if bool(retval) == False:
                self.dbusController.UpdateDetectFinishedNext(False,[''],get_error_code_from_enum(ERROR_UPDATE_SOURCE_FAILED),\
                                                            get_error_string_from_enum(ERROR_UPDATE_SOURCE_FAILED),\
                                                            get_error_description_from_enum(ERROR_UPDATE_SOURCE_FAILED))
            else:
                self.start_update_backend()
        def _update_important_error(retval):
            logging.error(str(retval))
            self.dbusController.UpdateDetectFinishedNext(False,[''],get_error_code_from_enum(ERROR_UPDATE_SOURCE_FAILED),\
            get_error_string_from_enum(ERROR_UPDATE_SOURCE_FAILED),\
                    get_error_description_from_enum(ERROR_UPDATE_SOURCE_FAILED))
        
        if self.options.no_update_source is False and self.is_openkylin == False:
            try:
                obj = self.bus.get_object('com.kylin.software.properties', '/com/kylin/software/properties')

                methods = dbus.Interface(obj, dbus_interface='org.freedesktop.DBus.Introspectable').Introspect()
                interface = dbus.Interface(obj, dbus_interface='com.kylin.software.properties.interface')
                if "checkUpdatesInStages" in methods:
                    self.dbusController.UpdateDetectStatusChanged(10,_("Updating Source Template")+"(checkUpdatesInStages)")
                    interface.checkUpdatesInStages(timeout=20,reply_handler=_update_important_reply,error_handler=_update_important_error)
                else:
                    self.dbusController.UpdateDetectStatusChanged(10,_("Updating Source Template")+"(updateSourceTemplate)")
                    interface.updateSourceTemplate(timeout=20,reply_handler=_update_important_reply,error_handler=_update_important_error)
            except Exception as e:
                logging.warning(str(e))
                raise UpdateBaseError(ERROR_UPDATE_SOURCE_FAILED)
            raise UpdateProgressExit()

    def _get_important_data(self):
        try:
            important_list = []
            if self.is_openkylin == False:
                with open(self.IMPORTANT_LIST_PATH, 'r') as f:
                    data = f.read()
                    important_list = data.split()
            else:
                important_list = [self.SYSTEM_UPDATE_GROUPS]

            #去除重复内容
            important_list = list(set(important_list))

            return important_list
        except Exception as e:
            logging.error(str(e))
            return []

    def _check_config_upgrade(self,cache):
        need_upgrade = False
        self_upgrade = []

        #config包
        for pkg_name in [self.GROUPS_PKG_NAME]:
            if pkg_name in cache:
                self_pkg = cache[pkg_name]
                if self_pkg.is_installed:
                    if self_pkg.is_upgradable:
                        logging.info("Check: groups JSON ConfigPkgs(%s) start upgrading From %s to %s...",pkg_name,\
                                        self_pkg.installed.source_version,self_pkg.candidate.source_version)
                        try:
                            self_pkg.mark_install(True, False, True)
                            self_upgrade.append(pkg_name)
                            need_upgrade = True
                        except SystemError:
                            logging.error("Check: mark %s to upgrade Failed...",pkg_name)
                            raise UpdateBaseError(ERROR_NOT_CONFIGPKG_DEPENDENCIES)
                    else:
                        # 已安装版本
                        cur_version = getattr(self_pkg.installed, "version", None)
                        sources_version = None
                        # 源中最高版本
                        for version_obj in self_pkg.versions:
                            if version_obj.downloadable == False:
                                continue
                            sources_version = version_obj.source_version
                            self_pkg.candidate = version_obj
                            break

                        if cur_version and sources_version:
                            # 当前版本大于源中的版本 将config包进行降级 
                            if apt_pkg.version_compare(cur_version,sources_version) > 0:
                                logging.info("Check: ConfigPkgs(%s) need to downgrade from %s to %s...",pkg_name,cur_version,sources_version)
                                self_pkg.mark_install(True, False, True)
                                self_upgrade.append(pkg_name)
                                need_upgrade = True
                        logging.info("Check: ConfigPkgs(%s:%s) No need to upgrade...",pkg_name,self_pkg.installed.source_version)
                else:
                    logging.info("Check: groups JSON ConfigPkgs(%s) start new installing...",pkg_name)
                    try:
                        self_pkg.mark_install(True, False, True)
                        self_upgrade.append(pkg_name)
                        need_upgrade = True
                    except SystemError:
                        logging.error("Check: mark %s to install Failed...",pkg_name)
                        raise UpdateBaseError(ERROR_NOT_CONFIGPKG_DEPENDENCIES)
            else:
                logging.error("Check: groups JSON ConfigPkgs(%s) is not in Cache...",pkg_name)
                raise UpdateBaseError(ERROR_NOT_GROUPS_CONFIG)

        if need_upgrade == True:
            self.dbusController.UpdateDetectStatusChanged(95,_("Priority Upgrade Package being updated"))
            self.start_selfupgrade(push_content=self_upgrade)
            raise UpdateProgressExit()

    def _check_failed_packages(self,cache):
        need_upgrade = False
        self_upgrade = []
        failed_packages = []

        for pkg in cache:
            if pkg.is_installed and pkg._pkg.current_state == apt_pkg.CURSTATE_HALF_CONFIGURED:
                failed_packages.append(pkg.name)
        
        logging.info("Check: Packages state and Exception pkgs:%r...",failed_packages)

        for pkg_name in failed_packages:
            self_pkg = cache[pkg_name]
            if self_pkg.is_installed and self_pkg.is_upgradable:
                try:
                    logging.info("Check: (%s) start fixing system From %s to %s...",pkg_name,\
                                    self_pkg.installed.source_version,self_pkg.candidate.source_version)
                    self_pkg.mark_install(True,False,True)
                    self_upgrade.append(pkg_name)
                    need_upgrade = True
                except SystemError:
                    logging.error("Check: mark %s to upgrade Failed...",pkg_name)
                    raise UpdateBaseError(ERROR_NOT_FIX_SYSTEM)
            else:
                logging.info("Check: (%s) Not to be installed or upgrade...",pkg_name)
                raise UpdateBaseError(ERROR_NOT_FIX_SYSTEM)
        
        if need_upgrade == True:
            self.dbusController.UpdateDetectStatusChanged(95,_("Repairing the system APT environment"))
            self.start_selfupgrade(push_content=self_upgrade)
            raise UpdateProgressExit()

    def _check_self_upgrade(self,cache):
        need_upgrade = False
        self_upgrade = []
        
        channel_config = json_config().getWithDefault("update_channel_upgrade",default = None)
        if channel_config == None:
            logging.warning("Json: update_channel_upgrade item is None...")
            upgrade_list =  [self.BACKEND_PKG_NAME,self.APTD_PKG_NAME,self.FRONTEND_PKG_NAME]
        else:
            upgrade_list = channel_config.get("upgrade_list",[self.BACKEND_PKG_NAME,self.APTD_PKG_NAME,self.FRONTEND_PKG_NAME])

        for pkg_name in upgrade_list:
            if pkg_name in cache:
                self_pkg = cache[pkg_name]
                if self_pkg.is_installed:
                    if self_pkg.is_upgradable:
                        logging.info("Check: (%s) will upgrading From %s to %s...",pkg_name,\
                                        self_pkg.installed.source_version,self_pkg.candidate.source_version)
                        try:
                            logging.info("Check: (%s) start upgrading From %s to %s...",pkg_name,\
                                            self_pkg.installed.source_version,self_pkg.candidate.source_version)
                            self_pkg.mark_install(True,False,True)
                            self_upgrade.append(pkg_name)
                            need_upgrade = True
                        except SystemError:
                            logging.error("Check: mark %s to upgrade Failed...",pkg_name)
                            raise UpdateBaseError(ERROR_NOT_SELFPKG_DEPENDENCIES)
                    else:
                        logging.info("Check: (%s:%s) No need to upgrade...",pkg_name,self_pkg.installed.source_version)
                else:
                    logging.info("Check: (%s) Not to be installed...",pkg_name)
            else:
                logging.error("Check: (%s) The upgrade package is not in Cache...",pkg_name)
        
        if need_upgrade == True:
            self.dbusController.UpdateDetectStatusChanged(95,_("Priority Upgrade Package being updated"))
            self.start_selfupgrade(push_content=self_upgrade)
            raise UpdateProgressExit()

    def _check_system_incomplete(self,cache):
        if cache.dpkgJournalDirty() == True:
            self.start_fix_incomplete()
            raise UpdateProgressExit()

    def _check_system_broken(self,cache):
        if cache.check_system_broken() == True:
            fix_backend = get_backend(self, InstallBackend.ACTION_CHECK_BROKEN)
            fix_backend.start()
            raise UpdateProgressExit()

    def _check_finished_status(self):
        try:
            finished_status = "reboot"
            with open("/tmp/monitor-dbus-upgrade.log", 'r') as f:
                text = f.read()
                if "member=PowerOff" in text:
                    finished_status = "shutdown"
            return finished_status
        except Exception as exc:
            logging.error(exc)
            return "reboot"

    def _setup_dbus(self,bus):
        # check if there is another g-a-i already and if not setup one
        # listening on dbus
        try:
            bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                            bus,
                                            do_not_queue=True)
            logging.info("Initiate dbus success ...")
            UpdateManagerDbusControllerUtils(self,bus,bus_name)
            return  UpdateManagerDbusController(self,bus,bus_name)
        except dbus.exceptions.NameExistsException:
            if self.options.replace is False:
                logging.critical("Another daemon is already running")
                sys.exit(1)
            logging.warning("Replacing already running daemon")
            
            retry_reboot_times = 0
            the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                            UPDATER_DBUS_PATH)
            the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                timeout=300)
            time.sleep(1)
            while True:
                retry_reboot_times = retry_reboot_times + 1
                #当重试次数超过5次时退出程序
                if retry_reboot_times > 5:
                    logging.critical("Reboot backend is Failed...")
                    sys.exit(1)
                try:
                    bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                                    bus,
                                                    do_not_queue=True)
                    logging.warning("Replacing already running daemon to Success...")
                    return  UpdateManagerDbusController(self,bus,bus_name)
                except dbus.exceptions.NameExistsException:
                    the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                                    UPDATER_DBUS_PATH)
                    the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                        timeout=300)
                    logging.error("Dbus has not withdrawn and retry reboot times:%d...",retry_reboot_times)
                    time.sleep(1)
    
    def _setup_dbus_utils(self,bus):
        # check if there is another g-a-i already and if not setup one
        # listening on dbus
        try:
            bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                            bus,
                                            do_not_queue=True)
            return  UpdateManagerDbusControllerUtils(self,bus,bus_name)
        except dbus.exceptions.NameExistsException:
            if self.options.replace is False:
                logging.critical("Another daemon is already running")
                sys.exit(1)
            logging.warning("Replacing already running daemon")
            
            retry_reboot_times = 0
            the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                            UPDATER_DBUS_PATH)
            the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                timeout=300)
            time.sleep(1)
            while True:
                retry_reboot_times = retry_reboot_times + 1
                #当重试次数超过5次时退出程序
                if retry_reboot_times > 5:
                    logging.critical("Reboot backend is Failed...")
                    sys.exit(1)
                try:
                    bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                                    bus,
                                                    do_not_queue=True)
                    logging.warning("Replacing already running daemon to Success...")
                    return  UpdateManagerDbusController(self,bus,bus_name)
                except dbus.exceptions.NameExistsException:
                    the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                                    UPDATER_DBUS_PATH)
                    the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                        timeout=300)
                    logging.error("Dbus has not withdrawn and retry reboot times:%d...",retry_reboot_times)
                    time.sleep(1)

    def start_back_upgrade(self, pkglist, mode=InstallBackend.MODE_BACKGROUND_UPGRADE):
        try:
            install_backend = get_backend(self, InstallBackend.ACTION_BACKGROUND_UPGRADE, mode)
            install_backend.start_alone(push_content = pkglist)
        except Exception as e:
            logging.error(str(e))

    def start_back_update_cache(self, action= InstallBackend.ACTION_BACKGROUND_UPGRADE, mode=InstallBackend.MODE_OTHER):
        try:
            install_backend = get_backend(self, action, mode)
            install_backend.start_alone()
        except Exception as e:
            logging.error(str(e))

    def start_deb_resolver(self, pkglist = []):
        try:
            install_backend = get_backend(self, InstallBackend.ACTION_INSTALL_DEB_RESOLVER)
            install_backend.start_alone(push_content = pkglist)
        except Exception as e:
            logging.error(str(e))

    # 进行本地deb包安装的操作
    # _check_local_dep : 是否查询本地依赖
    # _auto_satisfy    : 是否通过网络下载依赖
    def start_deb_install(self, deb_path = "", _check_local_dep = False, _auto_satisfy = False, source = '', sender=None):
        header = ''
        desc = ''
        absolute_path, debname = os.path.split(deb_path)
        self.sqlite3_server.deb_metadata.update({"current_install_debfile":deb_path})
        self.sqlite3_server.deb_metadata.update({"absolute_path":absolute_path})
        self.sqlite3_server.deb_metadata.update({"source":str(source)})
        
        self.deb_obj = {}
        UpdateMsg = {}
        try:
            if not os.path.isfile(deb_path):
                logging.info("No such file or directory: %s.",deb_path)
                self.dbusController.InstalldebFinished(False,'No such file or directory .','')
                return
            # 验签提权
            sender_name = get_proc_from_dbus_name(sender)
            caller = get_caller_from_enum(sender_name)
            caller_trans = get_source_name_from_enum(sender_name)
            if not deb_verify(deb_path): #验签失败，提权
                self.dbusController.InstalldebFinished(False,_("The package is unsigned, refuses to install."),'')
                return 
            self.deb_obj.update({"debname":str(debname)})
            self.deb_obj.update({"old_version":""})
            self.deb_obj.update({"source":str(source)})
            UpdateMsg.update({"source":str(self.deb_obj.get("source","kylin-system-updater"))})
            
            deb_cache = MyCache(None)

            ins = self._suit_install_mode(deb_path, deb_cache)
            UpdateMsg.update({"appname":str(self.debName)})
            self.sqlite3_server.deb_metadata.update({"debname":str(self.debName)})
            self.sqlite3_server.deb_metadata['deblist'].append(str(self.debName))
            UpdateMsg.update({"new_version":str(self.debVersion)})

            status,error_string = ApplicationSecInstall(str(self.debName))
            if not status:
                self.dbusController.InstalldebFinished(False,error_string,'')
                return 

            if self._is_broken > 0 or not self.cacheSatisfy:
            # 走 dpkg 安装流程，说明本地apt环境已经损坏 or dep not satisfied
                dep_satisfy, header, desc = self._deb_install(deb_cache, deb_path, _check_local_dep, ins)
                if dep_satisfy:
                    self.dbusController.InstalldebFinished(True, header, desc)
                    UpdateMsg.update({"status":"success"})
                    UpdateMsg.update({"errorCode":" "})
                else:
                    self.dbusController.InstalldebFinished(False, header, desc)
                    UpdateMsg.update({"status":"failed"})
                    UpdateMsg.update({"errorCode":str(header+"， "+desc)})
                #dpkg发送数据
                if True:
                    UpdateMsg.update({"old_version":str(self.deb_obj.get("old_version","None"))})
                    self.collector.Upgrade_Process_Msg(InstallBackend.ACTION_INSTALL_DEB, UpdateMsg.copy())
                    self.deb_obj = {}
            else:
            # apt 安装流程
                dep_satisfy, header, desc = self._attempt_depends(deb_cache, deb_path, _check_local_dep,_auto_satisfy, ins)
                if dep_satisfy:
                    install_backend = get_backend(self, InstallBackend.ACTION_INSTALL_DEB)
                    logging.info("source name: %s.", source)
                    install_backend.start_alone(push_content=deb_path,_is_install=_auto_satisfy,caller=source)
                else:
                    self.dbusController.InstalldebFinished(False, header, desc)
        except UpdateBaseError as excep:
            self.dbusController.InstalldebFinished(False,excep.header,excep.desc)
            UpdateMsg.update({"old_version":str(self.deb_obj.get("old_version","None"))})
            UpdateMsg.update({"errorCode":str(excep.header+"， "+excep.desc)})
            UpdateMsg.update({"status":"failed"})
            self.collector.Upgrade_Process_Msg(InstallBackend.ACTION_INSTALL_DEB, UpdateMsg.copy())
        except Exception as e:
            logging.info(str(e))
            traceback.print_exc()
            self.dbusController.InstalldebFinished(False, str(e), desc)
            UpdateMsg.update({"old_version":str(self.deb_obj.get("old_version","None"))})
            UpdateMsg.update({"errorCode":str(e)})
            UpdateMsg.update({"status":"failed"})
            self.collector.Upgrade_Process_Msg(InstallBackend.ACTION_INSTALL_DEB, UpdateMsg.copy())

    #进行删除的操作
    def start_purge_pkgs(self,pkgs_list):
        try:
            # 是否有破损的包
            deb_cache = Cache()
            broken_count = deb_cache._depcache.broken_count
            if broken_count > 0:
                _success,header,desc = self._dpkg_purge_pkgs(pkgs_list)
                if _success == True:
                    logging.info(header)
                    self.dbusController.PurgePackagesFinished(_success,'',desc)
                    self.dbusController.PurgePackagesFinishedWithPkgname(_success,'',desc,",".join(pkgs_list))
                else:
                    self.dbusController.PurgePackagesFinished(_success,header,desc)
                    self.dbusController.PurgePackagesFinishedWithPkgname(_success,header,desc,",".join(pkgs_list))
            else:
                purge_backend = get_backend(self, InstallBackend.ACTION_REMOVE_PACKAGES)
                purge_backend.start(push_content = pkgs_list)
            
            deb_cache.close()
        except Exception as e:
            logging.error(e)
            traceback.print_exc()

    def _dpkg_purge_pkgs(self,pkgs_list):
        success = False
        pkg_list_str = " ".join(pkgs_list)
        args = ["dpkg", "-r"]
        args.extend([pkg_list_str])

        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
        success = p.returncode == 0
        return success,p.stdout,''

    def _reload_options_config(self):
        apt_pkg.config["DPkg::Options::"] = "--force-confold"
        options_new = list(set(apt_pkg.config.value_list("DPkg::Options")))
        for option in ("--force-confnew","--force-confdef"):
            if option in options_new:
                options_new.remove(option)
        apt_pkg.config.clear("DPkg::Options")
        for option in options_new:
            apt_pkg.config["DPkg::Options::"] = option
        if apt_pkg.config.find_b("APT::Install-Recommends",False) == True:
            apt_pkg.config.clear("APT::Install-Recommends")
        if apt_pkg.config.find_b("APT::Install-Suggests",False) == True:
            apt_pkg.config.clear("APT::Install-Suggests")

    def check_config_aptdeamon(self):
    #检查是否需要重新启动aptdeamon 目前需要重启的有限速功能
        if self.init_config_aptdeamon == True:
            self.init_config_aptdeamon = False
            self.dbusController.set_aptdeamon_environ("init","config")

    # 是否查找本地依赖
    def _attempt_depends(self,deb_cache, deb_path,_check_local_dep,_auto_satisfy, _install):  
        depends_list = []
        depends_pkg = []
        satisfy_list = []
        depends_count = 0
        _local_satisfy = False
        error_string = ''
        error_desc = ''
        absolute_path, debname = os.path.split(deb_path)
        # 依赖不满足的情况
        if len(_install) > 0:
            if _check_local_dep: #查找本地
                # 需要查找本地依赖
                if len(_install) > 0:
                    for pkg in deb_cache:
                        if pkg.marked_install and pkg.name != str(debname.split("_")[0]):
                            depends_pkg.append(pkg)
                        elif pkg.marked_upgrade and pkg.name != str(debname.split("_")[0]):
                            depends_pkg.append(pkg)
                if len(depends_pkg)>0: #查找本地deb包
                    depends_list = [debfile for debfile in os.listdir(absolute_path) if debfile.endswith(".deb")]
                    for depends in depends_pkg:
                        for debfile in depends_list:
                            if depends.name in debfile and depends.candidate.version in debfile:
                                #FIXME:检查depends包的合法性
                                depends_count += 1
                                satisfy_list.append(debfile)
                    if depends_count < len(depends_pkg):
                        #本地依赖不满足 判断源下载
                        if _auto_satisfy:
                            _local_satisfy = True
                        elif not _auto_satisfy:
                            _local_satisfy = False
                            error_string = str(debname.split("_")[0])+_("dependency is not satisfied")
                            error_desc   = ",".join(_install)
                            logging.error(error_string+ error_desc)
                    else:
                        #将应用包与依赖包拷贝至archive目录安装
                        try:
                            if debname not in satisfy_list:
                                satisfy_list.append(debname)
                            for satisfy in satisfy_list:
                                shutil.copy(os.path.join(absolute_path,satisfy),"/var/cache/apt/archives/")
                                logging.info("move debpkg: %s",satisfy)
                            _local_satisfy = True
                        except Exception as e:
                            logging.info(str(e))
                return _local_satisfy,error_string,error_desc
            elif not _check_local_dep and _auto_satisfy:
                _local_satisfy = True
                if _install:
                    error_string = str(debname.split("_")[0])+_("dependency is not satisfied will download")
                    error_desc   = ",".join(_install)
                    logging.error(error_string+error_desc)
                    return _local_satisfy,error_string,error_desc
            elif not _check_local_dep and not _auto_satisfy:
                _local_satisfy = False
                if _install:
                    error_string = str(debname.split("_")[0])+_("dependency is not satisfied")
                    error_desc   = ",".join(_install)
                    logging.error(error_string+error_desc)
                    return _local_satisfy,error_string,error_desc
        # 依赖满足
        else:
            _local_satisfy = True
            error_string = ''
            error_desc=''
            return _local_satisfy,error_string,error_desc
    
    def _deb_install(self, deb_cache, deb_path, _check_local_dep, _install):
        local_depends_deb_list = []
        satisfy_list = []
        noSatisfyList = []
        self.depends_count = 0
        error_string = ''
        header = ''
        desc = ''
        absolute_path, debname = os.path.split(deb_path)
        debPackage = DebPackage(deb_path, deb_cache)
        self.absolute_path = absolute_path

        if _check_local_dep: # 需要查找本地依赖
            if len(self.noSatisfyList) > 0:
                for nS in self.noSatisfyList:
                    if str(nS[0][0]) not in noSatisfyList:
                        noSatisfyList.append(str(nS[0][0]))

            if len(noSatisfyList) > 0: #查找本地deb包
                local_depends_deb_list = [debfile for debfile in os.listdir(absolute_path) if debfile.endswith(".deb")]

                for nS in self.noSatisfyList:
                    for debfile in local_depends_deb_list:
                        # if "%3a" in debfile:
                        #     debfile=debfile.replace("%3a",":")
                        dep = DebPackage(os.path.join(absolute_path,debfile), deb_cache)
                        
                        if str(nS[0][0]) == dep.pkgname:# and debfile not in satisfy_list:
                            is_resolve = self._depend_resolver(dep, debfile, deb_cache, local_depends_deb_list, satisfy_list)
                            if is_resolve and debfile not in satisfy_list:
                                self.depends_count += 1
                                satisfy_list.append(debfile)

                if self.depends_count < len(noSatisfyList): #本地依赖不满足
                    error_string = str(debPackage.pkgname)+_("dependency is not satisfied")+", ".join(noSatisfyList)
                    logging.error(error_string)
                    header = error_string
                    return False,header, desc
                else: #安装依赖包
                    if debname not in satisfy_list:
                        satisfy_list.append(debname)
                    for satisfy in satisfy_list:
                        try:
                            deb = DebPackage(os.path.join(absolute_path,satisfy))
                            # iprogress = LogInstallProgress(os.path.join(absolute_path,satisfy))
                            ret = deb.install() #install_progress=iprogress
                            if ret != 0:
                                return False, "软件包安装失败!", desc
                        except Exception as e:
                            logging.error(e)
                            return False, str(e), desc
                    return True, header, desc
            else:
                try:
                    deb = DebPackage(deb_path)
                    # iprogress = LogInstallProgress(os.path.join(absolute_path,satisfy))
                    ret = deb.install() #install_progress=iprogress
                    if ret != 0:
                        return False, "软件包安装失败!", desc
                except Exception as e:
                    logging.error(e)
                    return False, str(e), desc
                return True, header, desc
        elif not _check_local_dep:
            error_string = str(debPackage.pkgname)+_("dependency is not satisfied")+":".join(_install)
            logging.error(error_string)
            header = error_string
            return False,header, desc

    def _depend_resolver(self, deb, debfile, deb_cache, local_depends_deb_list, satisfy_list):
        depends_cnt = 0
        is_resolve = False
        depends = deb.depends
        noSatisfyList = self._gen_noSatisfyList(depends, deb_cache)
        if len(noSatisfyList) > 0:
            for nS in noSatisfyList:
                for debfile in local_depends_deb_list:
                    if "%3a" in debfile:
                        debfile=debfile.replace("%3a",":")
                    dep = DebPackage(os.path.join(self.absolute_path,debfile), deb_cache)

                    if str(nS[0][0]) == dep.pkgname:
                        is_resolve = self._depend_resolver(dep, debfile, deb_cache, local_depends_deb_list, satisfy_list)
                        if is_resolve and debfile not in satisfy_list:
                            depends_cnt += 1
                            satisfy_list.append(debfile)
            if depends_cnt < len(noSatisfyList):
                is_resolve = False
        else:
                is_resolve = True
        return is_resolve

    def _suit_install_mode(self, deb_path, deb_cache):
        self._is_broken = False
        self.cacheSatisfy = False

        # 检查本地破损
        try:
            broken_count = deb_cache._depcache.broken_count
            logging.info("Install deb_package: check broken: %d. ", broken_count)
            self.debVersion = 0
            debPackage = DebPackage(deb_path, deb_cache)
            debPackage.check(allow_downgrade=True)
            self.debVersion = debPackage._sections["Version"]
            self.debName = debPackage.pkgname
            logging.info("Install deb_package: required changes")
            (install, remove, unauth) = debPackage.required_changes    # need in cach
            if broken_count > 0:
                self._is_broken = True
            else :
                self._is_broken = False
            if debPackage.pkgname in deb_cache:
                pkg = deb_cache[debPackage.pkgname]
                if pkg.is_installed:
                    self.deb_obj.update({"old_version":str(pkg.installed.version)})
        except apt_pkg.Error as e:
            logging.error(str(e))
            if "E:read" in str(e):
                raise UpdateBaseError(ERROR_READ_LOCAL_DEB)
            else:
                raise UpdateBaseError(ERROR_LOCAL_DEB_FORMAT)
        except Exception as e:
            logging.error(str(e))
            raise UpdateBaseError(ERROR_INSTALL_DEB_BASE)
        
        # 不满足的依赖列表
        depends = debPackage.depends
        self.noSatisfyList = self._gen_noSatisfyList(depends, deb_cache)
        _list = []
        # cache是否满足
        if len(install) == 0 and len(self.noSatisfyList) == 0:
            self.cacheSatisfy = True
        else:
            for ns in self.noSatisfyList:
                for or_group in ns:
                    for pkg in install:
                        if pkg == or_group[0]:
                            _list.append(ns)
            if len(_list) == len(self.noSatisfyList):
                self.cacheSatisfy = True
            else:
                self.cacheSatisfy = False
        logging.info("Cache satisfy is %r.",self.cacheSatisfy)
        return install
    
    def _gen_noSatisfyList(self, depends, deb_cache):
        _noSatisfyList = []
        _group_satify=False
        providers = []
        for or_group in depends:
            for deb_info in or_group:
                debname,ver,oper = deb_info
                if ":" in debname:
                    debname=debname.split(":")[0]
                try:
                    if _group_satify:
                        logging.warning("Group depend satify, ignore %s.", debname)
                        continue
                    if debname not in deb_cache:
                        if  deb_cache.is_virtual_package(debname):
                            logging.warning("The cache has no package named '%s', this is a virtual package...", debname)
                            providers = deb_cache.get_providing_packages(debname)
                        # hear!
                        if len(providers) < 1:
                            logging.error("Can not found depend %s.", debname)
                            continue
                        depname = providers[0].name
                        pkg = deb_cache[depname]
                        cand = deb_cache._depcache.get_candidate_ver(pkg._pkg)
                        logging.warning("Found virtual package %s, providers name: %s.", debname,depname)
                        if not cand:
                            _group_satify = False
                        if not apt_pkg.check_dep(cand.ver_str, oper, ver):
                            _group_satify = False
                        else:
                            _group_satify = True
                    else:
                        pkg = deb_cache[debname]
                        if (apt_pkg.check_dep(pkg.candidate.source_version, oper, ver)) or (pkg.installed and apt_pkg.check_dep(pkg.installed.version, oper, ver)) or (not pkg.installed and apt_pkg.check_dep(pkg.candidate.version, oper, ver)) and _group_satify == False:
                            # if not apt_pkg.check_dep(cand, oper, ver): #no candidate
                            _group_satify = True
                except Exception as e:
                    logging.error(str(e))
            if _group_satify == False and or_group not in _noSatisfyList:
                _noSatisfyList.append(or_group)
            _group_satify = False
        return _noSatisfyList


class AptP2pConfigManager():
    APT_P2P_FILE = "/etc/apt-p2p/apt-p2p.conf"
    HEADER_DSC = "apt-p2p config(/etc/apt-p2p/apt-p2p.conf) is not exists..."
    def __init__(self):
        if os.path.exists(self.APT_P2P_FILE):
            self.p2pConfigs = UpgradeConfig(datadir = "/etc/apt-p2p/", name = "apt-p2p.conf")
        else:
            self.p2pConfigs = None
    
    def get_bootstrap(self):
        if self.p2pConfigs == None:
            return self.HEADER_DSC

        return self.p2pConfigs.getWithDefault("apt_p2p_Khashmir", "BOOTSTRAP", "Failed")

    def set_bootstrap(self,value):
        if self.p2pConfigs == None:
            return self.HEADER_DSC
        self.p2pConfigs.setValue("apt_p2p_Khashmir","BOOTSTRAP",str(value))

class DelayShutdownLock():
    # 禁止关机锁文件路径
    FILELOCK_PATH = "/tmp/lock/"
    SHUTDOWN_BLOCK_FILELOCK = "kylin-update.lock"
    
    def __init__(self,parent):
        self.inhibit_lock = None
        self.pidfile = None
        self.parent = parent
        self.bus = self.parent.bus

    #安装时禁止关机 进行加锁
    def lock(self, caller='kylin-system-updater',mode='delay'):
        """
        Send a dbus signal to logind to not suspend the system, it will be
        released when the return value drops out of scope
        """
        if self.inhibit_lock == None:
            try:
                from gi.repository import Gio, GLib
                logging.info("Start acquiring Shutdown lock(%s)...",mode)
                connection = Gio.bus_get_sync(Gio.BusType.SYSTEM)
                
                var, fdlist = connection.call_with_unix_fd_list_sync(
                    'org.freedesktop.login1', '/org/freedesktop/login1',
                    'org.freedesktop.login1.Manager', 'Inhibit',
                    GLib.Variant('(ssss)',
                                ('shutdown:sleep',
                                caller, _("Installing or removing software packages."),
                                mode)),
                    None, 0, -1, None, None)
                self.inhibit_lock = Gio.UnixInputStream(fd=fdlist.steal_fds()[var[0]])
                logging.info("Shutdown Has been locked...")
                return True
            except Exception as e:
                logging.error(e)
                return False
        else:
            logging.info("No need to acquired lock and Shutdown Has been locked...")
            return False

    #解锁禁止关机 
    def unlock(self):
        try:
            if self.inhibit_lock != None:
                self.inhibit_lock.close()
                self.inhibit_lock = None
                logging.info("Shutdown Has been unlocked...")
            else:
                logging.info("Not locked and Quitting ...")
        except Exception as e:
            self.inhibit_lock = None
            logging.error("unlock failed." + str(e))

    def connect_logind(self):
        try:
            logind_proxy = self.bus.get_object('org.freedesktop.login1', '/org/freedesktop/login1')
            def prepare_for_shutdown_handler(active):
                """ Handle PrepareForShutdown() """
                if not active:
                    logging.warning("PrepareForShutdown(false) received, "
                                    "this should not happen")

                logging.info("Prepare For Shutdown arrived, starting final iterations....")
                
                self.parent.start_poweroff('user')

            logind_proxy.connect_to_signal(
                    "PrepareForShutdown", prepare_for_shutdown_handler)

            self._start_monitor_device()
        except Exception as exc:
            logging.error(exc)

    def _start_monitor_device(self): 
        try:
            args = ["/usr/share/kylin-system-updater/monitor-reboot"]
            p = subprocess.run(args,stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
            logging.info("Start monitor device Trigger mothod for Reboot and Poweroff...")
        except Exception as exc:
            logging.error(exc)
