# UpdateManager.py
# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-
import os
import sys
import time
import dbus
import logging
import dbus.service
import traceback
import subprocess
import apt_pkg
from gettext import gettext as _
from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import GLib
DBusGMainLoop(set_as_default=True)

from .UpgradeStrategiesDbus import UpgradeStrategiesDbusController,UPDATER_DBUS_INTERFACE,UPDATER_DBUS_PATH,UPDATER_DBUS_SERVICE
from .Core.Database import Sqlite3Server
from .Core.loop import mainloop
from .Core.MyCache import MyCache
from .backend import InstallBackend,get_backend
from .Core.DataAcquisition import UpdateMsgCollector
from SystemUpdater.Core.LocalUpgradeData import LocalUpgradeDataList
from SystemUpdater.Core.UpdaterConfigParser import UpgradeConfig

STRATEGY_IDLE_INTERVAL = 2*60
STRATEGY_IDLE_TIMEOUT = 6*60
class UpgradeStrategies():
    def __init__(self,options):
        try:
            self.options = options
            self.cache = None
            self.aptd_lang_switch = False
            self.now_working = InstallBackend.ACTION_DEFUALT_STATUS

            #dbus
            self.dbusController = self._setup_dbus()

            self.bus = dbus.SystemBus()
            #config
            self.uuconfigs = UpgradeConfig(datadir = "/var/lib/unattended-upgrades/", name = "unattended-upgrades-policy.conf")
            self.configs_uncover = UpgradeConfig(defaults_dir="system-updater-defaults.conf")
            self.configs_cover = UpgradeConfig(name = "system-updater-coverable.conf")
            self.sqlite3_server = Sqlite3Server(self, _no_DataMigration=True)
            #策略配置接口的超时退出机制
            self.strategy_timestamp = 0
            # GLib.timeout_add_seconds(STRATEGY_IDLE_INTERVAL,
            #     self._check_strategy_inactivity)
            
            #数据采集器
            self.collector = UpdateMsgCollector(self)
            self.local_meta = LocalUpgradeDataList()
            self.delay_shutdown = DelayShutdownLock(self)

            self.refresh_cache()
        except Exception as e:
            logging.error(e)
            traceback.print_exc()
        
    def refresh_cache(self):
        try:
            self._reload_options_config()
            if self.cache is None:
                self.cache = MyCache(None)
            else:
                self.cache.open(None)
                self.cache._initDepCache()
            logging.info("Refresh: Cache for System Update Successfully...")
        # except AssertionError:
        #     raise UpdateBaseError(ERROR_SOFTWARE_INDEX_RROKEN)
        # except SystemError as e:
        #     logging.error(str(e))
        #     raise UpdateBaseError(ERROR_NOT_INIT_PACKAGESINFIO)
        except Exception as e:
            logging.error(e)
    
    def run(self):
        """Start the daemon and listen for calls."""
        logging.info("Waiting for calls...")
        try:
            mainloop.run()
        except KeyboardInterrupt:
            self.dbusController.Quit(None)

    def _setup_dbus(self):
        # check if there is another g-a-i already and if not setup one
        # listening on dbus
        bus = dbus.SystemBus()
        try:
            bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                            bus,
                                            do_not_queue=True)
            logging.info("Initiate dbus success ...")
            return  UpgradeStrategiesDbusController(self, bus_name)
        except dbus.exceptions.NameExistsException:
            if self.options.replace is False:
                logging.critical("Another daemon is already running")
                sys.exit(1)
            logging.warning("Replacing already running daemon")
            
            retry_reboot_times = 0
            the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                            UPDATER_DBUS_PATH)
            the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                timeout=300)
            time.sleep(1)
            while True:
                retry_reboot_times = retry_reboot_times + 1
                #当重试次数超过5次时退出程序
                if retry_reboot_times > 5:
                    logging.critical("Reboot backend is Failed...")
                    sys.exit(1)
                try:
                    bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                                    bus,
                                                    do_not_queue=True)
                    logging.warning("Replacing already running daemon to Success...")
                    return  UpgradeStrategiesDbusController(self, bus_name)
                except dbus.exceptions.NameExistsException:
                    the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                                    UPDATER_DBUS_PATH)
                    the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                        timeout=300)
                    logging.error("Dbus has not withdrawn and retry reboot times:%d...",retry_reboot_times)
                    time.sleep(1)
    
    def _check_strategy_inactivity(self):
        logging.info("Checking for inactivity in Strategies daemon ...")
        timestamp = self.strategy_timestamp
        if timestamp == 0:
            self.strategy_timestamp = time.time()
            return True
        #超时退出
        if self.strategy_timestamp != 0 and  time.time() - self.strategy_timestamp > STRATEGY_IDLE_TIMEOUT:
            logging.warning("Quitting due to inactivity")
            self.dbusController.Quit(None)
            return False
        return True

    def start_update_backend(self):
        # self.check_config_aptdeamon()
        update_backend = get_backend(self, InstallBackend.ACTION_UPDATE,InstallBackend.MODE_UPDATE_CACHE)
        update_backend.start()

    def _reload_options_config(self):
        apt_pkg.config["DPkg::Options::"] = "--force-confold"
        options_new = list(set(apt_pkg.config.value_list("DPkg::Options")))
        for option in ("--force-confnew","--force-confdef"):
            if option in options_new:
                options_new.remove(option)
        apt_pkg.config.clear("DPkg::Options")
        for option in options_new:
            apt_pkg.config["DPkg::Options::"] = option
        if apt_pkg.config.find_b("APT::Install-Recommends",False) == True:
            apt_pkg.config.clear("APT::Install-Recommends")
        if apt_pkg.config.find_b("APT::Install-Suggests",False) == True:
            apt_pkg.config.clear("APT::Install-Suggests")

class DelayShutdownLock():
    # 禁止关机锁文件路径
    FILELOCK_PATH = "/tmp/lock/"
    SHUTDOWN_BLOCK_FILELOCK = "kylin-update.lock"
    
    def __init__(self,parent):
        self.inhibit_lock = None
        self.pidfile = None
        self.parent = parent
        self.bus = self.parent.bus

    #安装时禁止关机 进行加锁
    def lock(self, caller='kylin-system-updater',mode='delay'):
        """
        Send a dbus signal to logind to not suspend the system, it will be
        released when the return value drops out of scope
        """
        if self.inhibit_lock == None:
            try:
                from gi.repository import Gio, GLib
                logging.info("Start acquiring Shutdown lock(%s)...",mode)
                connection = Gio.bus_get_sync(Gio.BusType.SYSTEM)
                
                var, fdlist = connection.call_with_unix_fd_list_sync(
                    'org.freedesktop.login1', '/org/freedesktop/login1',
                    'org.freedesktop.login1.Manager', 'Inhibit',
                    GLib.Variant('(ssss)',
                                ('shutdown:sleep',
                                caller, _("Installing or removing software packages."),
                                mode)),
                    None, 0, -1, None, None)
                self.inhibit_lock = Gio.UnixInputStream(fd=fdlist.steal_fds()[var[0]])
                logging.info("Shutdown Has been locked...")
                return True
            except Exception as e:
                logging.error(e)
                return False
        else:
            logging.info("No need to acquired lock and Shutdown Has been locked...")
            return False

    #解锁禁止关机 
    def unlock(self):
        try:
            if self.inhibit_lock != None:
                self.inhibit_lock.close()
                self.inhibit_lock = None
                logging.info("Shutdown Has been unlocked...")
            else:
                logging.info("Not locked and Quitting ...")
        except Exception as e:
            self.inhibit_lock = None
            logging.error("unlock failed." + str(e))

    def connect_logind(self):
        try:
            logind_proxy = self.bus.get_object('org.freedesktop.login1', '/org/freedesktop/login1')
            def prepare_for_shutdown_handler(active):
                """ Handle PrepareForShutdown() """
                if not active:
                    logging.warning("PrepareForShutdown(false) received, "
                                    "this should not happen")

                logging.info("Prepare For Shutdown arrived, starting final iterations....")
                
                self.parent.start_poweroff('user')

            logind_proxy.connect_to_signal(
                    "PrepareForShutdown", prepare_for_shutdown_handler)

            self._start_monitor_device()
        except Exception as exc:
            logging.error(exc)

    def _start_monitor_device(self): 
        try:
            args = ["/usr/share/kylin-system-updater/monitor-reboot"]
            p = subprocess.run(args,stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
            logging.info("Start monitor device Trigger mothod for Reboot and Poweroff...")
        except Exception as exc:
            logging.error(exc)
